// Pràctica 7 OPENGL: Múltiples càmeres

#include <math.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <ocell/core.h>
#include <ocell/super_cast.h>

enum {
	W_INIT = 250
};

// Grandària del volum de retall
static GLfloat volx = W_INIT / 2;
static GLfloat voly = W_INIT / 2;
static GLfloat volz1 = 500;
static GLfloat volz2 = 1000;

// Coordenades per la càmera mòbil
static GLdouble eyeX = 0;
static GLdouble eyeY = 0;
static GLdouble eyeZ = 0;

// Angle de vista per la càmera mòbil
static GLdouble angleview = 0;

static _Bool ortho_projection;

#define CAMERA_CAST(p, t) SUPER_CAST(p, t, camera, process, p)

typedef struct camera camera;
typedef struct {
	GLfloat radius;
	GLfloat length;
	GLUquadricObj *quad;
	GLUquadricObj *quad_base;
} camera_private;

#define CUBE_CAST(p, t) \
	SUPER_CAST(p, t, cube, object_3d, \
		OBJECT_3D_CAST((struct object_3d *) p, t))

typedef struct cube cube;
typedef struct {
	GLsizei size;
} cube_private;

#define IG_ROOM_CAST(p, t) SUPER_CAST(p, t, ig_room, process, p)

typedef struct ig_room ig_room;
typedef struct {
	// Buffer de profunditat
	_Bool depth;
	// Ombra
	_Bool shade;
} ig_room_private;

#define KEY_HANDLER_CAST(p, t) SUPER_CAST(p, t, key_handler, process, p)

typedef struct key_handler key_handler;

#define LIGHT_BULB_CAST(p, t) \
	SUPER_CAST(p, t, light_bulb, object_3d, \
		OBJECT_3D_CAST((struct object_3d *) p, t))

typedef struct light_bulb light_bulb;
typedef struct {
	_Bool on;
	GLsizei radius;
	GLUquadricObj *quad;
} light_bulb_private;

#define OBJECT_3D_CAST(p, t) SUPER_CAST(p, t, object_3d, process, p)

typedef struct object_3d object_3d;

#define PYRAMID_CAST(p, t) \
	SUPER_CAST(p, t, pyramid, object_3d, \
		OBJECT_3D_CAST((struct object_3d *) p, t))

typedef struct pyramid pyramid;
typedef struct {
	GLsizei size;
} pyramid_private;

#define ROOM_CAST(p, t) \
	SUPER_CAST(p, t, room, object_3d, \
		OBJECT_3D_CAST((struct object_3d *) p, t))

typedef struct room room;

#define SPHERE_CAST(p, t) \
	SUPER_CAST(p, t, sphere, object_3d, \
		OBJECT_3D_CAST((struct object_3d *) p, t))

typedef struct sphere sphere;
typedef struct {
	GLfloat spX;
	GLfloat spY;
	GLfloat spZ;
	GLsizei radius;
	GLfloat stepX;
	GLfloat stepY;
	GLfloat stepZ;
	GLUquadricObj *quad;
} sphere_private;

static unsigned char camera_active;
static size_t private_offset;

static void camera_on_kill(process *p);
static void camera_render(process *p);
static void camera_run(process *p);
static size_t camera_sizeof(void);
static cube_private *cube_get_private(cube *c);
static void cube_render(process *p);
static size_t cube_sizeof(void);
static void ig_room_render(process *p);
static void ig_room_run(process *p);
static size_t ig_room_sizeof(void);
static key_handler *key_handler_create(process *father);
static void key_handler_run(process *p);
static size_t key_handler_sizeof(void);
static light_bulb *light_bulb_create(process *father);
static light_bulb_private *light_bulb_get_private(light_bulb *l);
static void light_bulb_on_kill(process *p);
static void light_bulb_render(process *p);
static void light_bulb_run(process *p);
static size_t light_bulb_sizeof(void);
static void object_3d_init(object_3d *o, process *father);
static pyramid *pyramid_create(process *father);
static pyramid_private *pyramid_get_private(pyramid *py);
static void pyramid_render(process *p);
static size_t pyramid_sizeof(void);
static room *room_create(process *father);
static void room_render(process *p);
static size_t room_sizeof(void);
static void solid_box(GLdouble bwidth, GLdouble bheight, GLdouble bdepth);
static void solid_cube(GLdouble csize);
static sphere *sphere_create(process *father);
static sphere_private *sphere_get_private(sphere *s);
static void sphere_on_kill(process *p);
static void sphere_render(process *p);
static void sphere_run(process *p);
static size_t sphere_sizeof(void);
static void wire_box(GLdouble bwidth, GLdouble bheight, GLdouble bdepth);
static void update_projection(void);

int main(void) {
	private_offset = process_sizeof();
	ig_room *i = alloc(ig_room_sizeof());
	process *p = IG_ROOM_CAST(i, process);
	program_init(p);
	p->render = ig_room_render;
	p->run = ig_room_run;
	run_program();
	return EXIT_SUCCESS;
}

static camera *camera_create(process *father) {
	camera *c = alloc(camera_sizeof());
	process *p = CAMERA_CAST(c, process);
	process_init(p, father);
	p->on_kill = camera_on_kill;
	p->render = camera_render;
	p->run = camera_run;
	p->run(p);
	return c;
}

static camera_private *camera_get_private(camera *c) {
	void *p = c;
	return p + private_offset;
}

static void camera_on_kill(process *p) {
	camera *c_p = (camera *) p;
	camera_private *c = camera_get_private(c_p);
	gluDeleteQuadric(c->quad);
	gluDeleteQuadric(c->quad_base);
}

static void camera_render(process *p) {
	glColor3f(1, 1, 0.5);
	glPushMatrix();

	camera *c_p = (camera *) p;
	camera_private *c = camera_get_private(c_p);
	// Posam el con segons els moviments que hem fet
	glTranslatef(eyeX + c->length * sin(angleview), eyeY,
		eyeZ - c->length * cos(angleview));
	glRotatef(-angleview * 180 / M_PI, 0, 1, 0);

	// Dibuixa un con
	gluCylinder(c->quad, c->radius, 0, c->length, 32, 1);
	gluDisk(c->quad_base, 0, c->radius, 32, 1);
	glPopMatrix();
}

static void camera_run(process *p) {
	if (p->state)
		goto *p->state;

	camera *c_p = (camera *) p;
	camera_private *c = camera_get_private(c_p);
	c->radius = 50;
	c->length = 100;
	c->quad = gluNewQuadric();
	c->quad_base = gluNewQuadric();
	gluQuadricOrientation(c->quad_base, GLU_INSIDE);

	for (;;) {
		static const GLdouble ANGLE_STEP = 3 * M_PI / 180;
		if (key(_LEFT)) {
			angleview -= ANGLE_STEP;
			if (angleview < 0)
				angleview += 2 * M_PI;
			update_projection();
		}
		if (key(_RIGHT)) {
			angleview += ANGLE_STEP;
			if (angleview > 2 * M_PI)
				angleview -= 2 * M_PI;
			update_projection();
		}

		static const int EYE_STEP = 10;
		if (key(_UP)) {
			eyeX += EYE_STEP * sin(angleview);
			eyeZ -= EYE_STEP * cos(angleview);
			update_projection();
		}
		if (key(_DOWN)) {
			eyeX -= EYE_STEP * sin(angleview);
			eyeZ += EYE_STEP * cos(angleview);
			update_projection();
		}

		if (key_down(_C)) {
			camera_active++;
			camera_active %= 4;
			p->graph = camera_active != 0 ? (struct graphic *) 1
				: NULL;
			update_projection();
		}

		FRAME(p);
	}
}

static size_t camera_sizeof(void) {
	static size_t size;
	if (!size)
		size = private_offset + sizeof(camera_private);
	return size;
}

static void change_size(int w, int h) {
	set_mode(w, h);
	update_projection();
}

static cube *cube_create(process *father) {
	cube *c = alloc(cube_sizeof());
	object_3d_init(CUBE_CAST(c, object_3d), father);
	process *p = CUBE_CAST(c, process);
	p->render = cube_render;
	cube_get_private(c)->size = 60;
	p->run(p);
	return c;
}

static cube_private *cube_get_private(cube *c) {
	void *p = c;
	return p + private_offset;
}

static void cube_render(process *p) {
	glColor3f(1, 1, 1);
	glPushMatrix();

	// El cub està al centre de l'habitació i semirotat
	glTranslatef(0, 0, -(volz1 + volz2) / 2);
	glRotatef(20, 1, 1, 1);

	cube *c = (cube *) p;
	solid_cube(cube_get_private(c)->size);
	glPopMatrix();
}

static size_t cube_sizeof(void) {
	static size_t size;
	if (!size)
		size = private_offset + sizeof(cube_private);
	return size;
}

static ig_room_private *ig_room_get_private(ig_room *i) {
	void *p = i;
	return p + private_offset;
}

static void ig_room_render(__attribute__((unused)) process *p) {
	glClearColor(0, 0, 1, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

static void ig_room_run(process *p) {
	ig_room *i_p = (ig_room *) p;
	ig_room_private *i = ig_room_get_private(i_p);
	if (p->state)
		goto *p->state;

	set_title("Etapa 7: Múltiples cámaras");
	set_fps(50, 0);
	resize_callback = change_size;
	change_size(W_INIT, W_INIT);

	i->depth = 1;
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHT0);
	// Ombrejat suau (SMOOTH)
	i->shade = 1;

	// Perspectiva
	ortho_projection = 0;

	// Components de la llum
	GLfloat ambient[] = {0.4, 0.4, 0.4, 1};
	GLfloat diffuse[] = {0.8, 0.8, 0.8, 1};
	GLfloat specular[] = {1, 1, 1, 1};

	// Propietats de la llum
	glEnable(GL_LIGHTING);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, specular);

	// Color dels materials
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

	p->z = 100;
	p->graph = (struct graphic *) 1;

	key_handler_create(p);
	room_create(p);
	cube_create(p);
	pyramid_create(p);
	sphere_create(p);
	light_bulb_create(p);
	camera_create(p);

	for (;;) {
		if (key_down(_O)) {
			ortho_projection = !ortho_projection;
			update_projection();
		}
		if (key_down(_P)) {
			i->depth = !i->depth;
			i->depth ? glEnable(GL_DEPTH_TEST)
				: glDisable(GL_DEPTH_TEST);
		}
		if (key_down(_S)) {
			i->shade = !i->shade;
			i->shade ? glShadeModel(GL_SMOOTH)
				: glShadeModel(GL_FLAT);
		}

		FRAME(p);
	}
	send_signal(p, S_KILL);
}

static size_t ig_room_sizeof(void) {
	static size_t size;
	if (!size)
		size = private_offset + sizeof(ig_room_private);
	return size;
}

static key_handler *key_handler_create(process *father) {
	key_handler *k = alloc(key_handler_sizeof());
	process *p = KEY_HANDLER_CAST(k, process);
	process_init(p, father);
	p->run = key_handler_run;
	p->run(p);
	return k;
}

static void key_handler_run(process *p) {
	if (p->state)
		goto *p->state;

	for (;;) {
		if ((key(_ALT) && key_down(_F4)) || key_down(_ESC)
			|| quit_requested) {
			let_me_alone(p);
			break;
		}
		if (key(_ALT) && key_down(_ENTER)) {
			full_screen = !full_screen;
			change_size(get_screen()->w, get_screen()->h);
		}

		FRAME(p);
	}
	send_signal(p, S_KILL);
}

static size_t key_handler_sizeof(void) {
	return process_sizeof();
}

static light_bulb *light_bulb_create(process *father) {
	light_bulb *l_p = alloc(light_bulb_sizeof());
	object_3d_init(LIGHT_BULB_CAST(l_p, object_3d), father);
	process *p = LIGHT_BULB_CAST(l_p, process);
	p->on_kill = light_bulb_on_kill;
	p->render = light_bulb_render;
	p->run = light_bulb_run;
	light_bulb_private *l = light_bulb_get_private(l_p);
	l->on = 1;
	l->radius = 5;
	l->quad = gluNewQuadric();
	p->run(p);
	return l_p;
}

static light_bulb_private *light_bulb_get_private(light_bulb *l) {
	void *p = l;
	return p + private_offset;
}

static void light_bulb_on_kill(process *p) {
	light_bulb *l = (light_bulb *) p;
	gluDeleteQuadric(light_bulb_get_private(l)->quad);
}

static void light_bulb_render(process *p) {
	glColor3f(1, 1, 1);
	glPushMatrix();

	light_bulb *l_p = (light_bulb *) p;
	light_bulb_private *l = light_bulb_get_private(l_p);
	// Si el llum està encès la bombeta emet llum
	if (l->on) {
		GLfloat LightOn[]  = {1, 1, 1, 1};
		glMaterialfv(GL_FRONT, GL_EMISSION, LightOn);
	}

	// La bombeta està centrada al sostre de l'habitació
	glTranslatef(0, voly - l->radius, -(volz1 + volz2) / 2);

	gluSphere(l->quad, l->radius, 32, 32);

	// La llum surt del centre de la bombeta
	GLfloat position[] = {0, 0, 0, 1};
	glLightfv(GL_LIGHT0, GL_POSITION, position);
	GLfloat LightOff[] = {0, 0, 0, 1};
	glMaterialfv(GL_FRONT, GL_EMISSION, LightOff);

	glPopMatrix();
}

static void light_bulb_run(process *p) {
	if (key_down(_L)) {
		light_bulb *l_p = (light_bulb *) p;
		light_bulb_private *l = light_bulb_get_private(l_p);
		l->on = !l->on;
		l->on ? glEnable(GL_LIGHT0) : glDisable(GL_LIGHT0);
	}
	frame_done(p);
}

static size_t light_bulb_sizeof(void) {
	static size_t size;
	if (!size)
		size = private_offset + sizeof(light_bulb_private);
	return size;
}

static int min_int(int a, int b) {
	return a < b ? a : b;
}

static void object_3d_init(object_3d *o, process *father) {
	process *p = OBJECT_3D_CAST(o, process);
	process_init(p, father);
	p->graph = (struct graphic *) 1;
}

static pyramid *pyramid_create(process *father) {
	pyramid *py = alloc(pyramid_sizeof());
	object_3d_init(PYRAMID_CAST(py, object_3d), father);
	process *p = PYRAMID_CAST(py, process);
	p->render = pyramid_render;
	pyramid_get_private(py)->size = 40;
	p->run(p);
	return py;
}

static pyramid_private *pyramid_get_private(pyramid *py) {
	void *p = py;
	return p + private_offset;
}

static void pyramid_render(process *p) {
	glPushMatrix();
	glColor3f(0, 1, 1);

	// La piràmide està al centre de l'habitació i semirotada (mirant cap
	// amunt)
	glTranslatef(0, 0, -(volz1 + volz2) / 2);
	glRotatef(20, 0, 1, 0);

	pyramid *py_p = (pyramid *) p;
	pyramid_private *py = pyramid_get_private(py_p);
	glBegin(GL_TRIANGLE_FAN);
	// L'alçada i cada costat de la base val 2 * size
	glNormal3f(0, 1, 0);
	glVertex3f(0, py->size, 0);
	static const float NVALUE = 1 / 1.732050808; // 1 / sqrtf(3)
	glNormal3f(-NVALUE, -NVALUE, -NVALUE);
	glVertex3f(-py->size, -py->size, -py->size);
	glNormal3f(NVALUE, -NVALUE, -NVALUE);
	glVertex3f(py->size, -py->size, -py->size);
	glNormal3f(NVALUE, -NVALUE, NVALUE);
	glVertex3f(py->size, -py->size, py->size);
	glNormal3f(-NVALUE, -NVALUE, NVALUE);
	glVertex3f(-py->size, -py->size, py->size);
	glNormal3f(-NVALUE, -NVALUE, -NVALUE);
	glVertex3f(-py->size, -py->size, -py->size);
	glEnd();

	glPopMatrix();
}

static size_t pyramid_sizeof(void) {
	static size_t size;
	if (!size)
		size = private_offset + sizeof(pyramid_private);
	return size;
}

static room *room_create(process *father) {
	room *r = alloc(room_sizeof());
	object_3d_init(ROOM_CAST(r, object_3d), father);
	process *p = ROOM_CAST(r, process);
	p->render = room_render;
	p->run(p);
	return r;
}

static void room_render(__attribute__((unused)) process *p) {
	glColor3f(0, 1, 0);
	glPushMatrix();

	// L'habitació està centrada al centre del volum de retall
	glTranslatef(0, 0, -(volz1 + volz2) / 2);

	wire_box(W_INIT, W_INIT, volz2 - volz1);

	glPopMatrix();
}

static size_t room_sizeof(void) {
	return process_sizeof();
}

static void solid_box(GLdouble bwidth, GLdouble bheight, GLdouble bdepth) {
	const GLdouble width = bwidth / 2;
	const GLdouble height = bheight / 2;
	const GLdouble depth = bdepth / 2;

	glBegin(GL_QUADS);

	glNormal3f(-1, 0, 0);
	glVertex3d(-width, -height, -depth);
	glVertex3d(-width, -height, depth);
	glVertex3d(-width, height, depth);
	glVertex3d(-width, height, -depth);

	glNormal3f(1, 0, 0);
	glVertex3d(width, -height, -depth);
	glVertex3d(width, -height, depth);
	glVertex3d(width, height, depth);
	glVertex3d(width, height, -depth);

	glNormal3f(0, -1, 0);
	glVertex3d(-width, -height, -depth);
	glVertex3d(-width, -height, depth);
	glVertex3d(width, -height, depth);
	glVertex3d(width, -height, -depth);

	glNormal3f(0, 1, 0);
	glVertex3d(-width, height, -depth);
	glVertex3d(-width, height, depth);
	glVertex3d(width, height, depth);
	glVertex3d(width, height, -depth);

	glNormal3f(0, 0, -1);
	glVertex3d(-width, -height, -depth);
	glVertex3d(-width, height, -depth);
	glVertex3d(width, height, -depth);
	glVertex3d(width, -height, -depth);

	glNormal3f(0, 0, 1);
	glVertex3d(-width, -height, depth);
	glVertex3d(-width, height, depth);
	glVertex3d(width, height, depth);
	glVertex3d(width, -height, depth);

	glEnd();
}

static void solid_cube(GLdouble csize) {
	solid_box(csize, csize, csize);
}

static sphere *sphere_create(process *father) {
	sphere *s_p = alloc(sphere_sizeof());
	object_3d_init(SPHERE_CAST(s_p, object_3d), father);
	process *p = SPHERE_CAST(s_p, process);
	p->on_kill = sphere_on_kill;
	p->render = sphere_render;
	p->run = sphere_run;
	sphere_private *s = sphere_get_private(s_p);
	s->spX = random_between(-25, 25);
	s->spY = random_between(-25, 25);
	s->spZ = -(volz2 + volz1) / 2 + random_between(-25, 25);
	s->radius = 40;
	s->stepX = 1;
	s->stepY = 1;
	s->stepZ = 4;
	s->quad = gluNewQuadric();
	p->run(p);
	return s_p;
}

static sphere_private *sphere_get_private(sphere *s) {
	void *p = s;
	return p + private_offset;
}

static void sphere_on_kill(process *p) {
	sphere *s = (sphere *) p;
	gluDeleteQuadric(sphere_get_private(s)->quad);
}

static void sphere_render(process *p) {
	glColor3f(1, 0, 0);
	glPushMatrix();

	sphere *s_p = (sphere *) p;
	sphere_private *s = sphere_get_private(s_p);
	glTranslatef(s->spX, s->spY, s->spZ);

	gluSphere(s->quad, s->radius, 32, 32);
	glPopMatrix();
}

static void sphere_run(process *p) {
	sphere *s_p = (sphere *) p;
	sphere_private *s = sphere_get_private(s_p);
	// Canviam de direcció si sortim del volum de retall
	if (s->spX > volx - s->radius || s->spX < -volx + s->radius)
		s->stepX = -s->stepX;
	if (s->spY > voly - s->radius || s->spY < -voly + s->radius)
		s->stepY = -s->stepY;
	if (-s->spZ > volz2 - s->radius || -s->spZ < volz1 + s->radius)
		s->stepZ = -s->stepZ;

	// Desplaçam les coordenades de l'esfera
	s->spX += s->stepX;
	s->spY += s->stepY;
	s->spZ += s->stepZ;

	frame_done(p);
}

static size_t sphere_sizeof(void) {
	static size_t size;
	if (!size)
		size = private_offset + sizeof(sphere_private);
	return size;
}

static void update_projection(void) {
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	GLfloat wsize = min_int(get_screen()->w, get_screen()->h);
	glScalef(wsize / get_screen()->w, wsize / get_screen()->h, 1);

	ortho_projection ? glOrtho(-volx, volx, -voly, voly, volz1, volz2) :
		gluPerspective(45.0, volx / voly, 1, 3000);

	// S'estableix la càmera activa
	switch (camera_active) {
	case 0:
		// càmera mòbil
		gluLookAt(eyeX, eyeY, eyeZ, eyeX + sin(angleview), 0,
			eyeZ - cos(angleview), 0, 1, 0);
		break;
	case 1:
		// càmera frontal
		gluLookAt(0, 0, 0, 0, 0, -1, 0, 1, 0);
		break;
	case 2:
		// càmera d'alçada
		gluLookAt(0, 800, -750, 0, 0, -750, 0, 0, -1);
		break;
	case 3:
		// càmera de perfil
		gluLookAt(800, 0, -750, 0, 0, -750, 0, 1, 0);
		break;
	}

	glMatrixMode(GL_MODELVIEW);
}

static void wire_box(GLdouble bwidth, GLdouble bheight, GLdouble bdepth) {
	const GLdouble width = bwidth / 2;
	const GLdouble height = bheight / 2;
	const GLdouble depth = bdepth / 2;

	glBegin(GL_LINE_LOOP);
	glNormal3f(-1, 0, 0);
	glVertex3d(-width, -height, -depth);
	glVertex3d(-width, -height, depth);
	glVertex3d(-width, height, depth);
	glVertex3d(-width, height, -depth);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glNormal3f(1, 0, 0);
	glVertex3d(width, -height, -depth);
	glVertex3d(width, -height, depth);
	glVertex3d(width, height, depth);
	glVertex3d(width, height, -depth);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glNormal3f(0, -1, 0);
	glVertex3d(-width, -height, -depth);
	glVertex3d(-width, -height, depth);
	glVertex3d(width, -height, depth);
	glVertex3d(width, -height, -depth);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glNormal3f(0, 1, 0);
	glVertex3d(-width, height, -depth);
	glVertex3d(-width, height, depth);
	glVertex3d(width, height, depth);
	glVertex3d(width, height, -depth);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glNormal3f(0, 0, -1);
	glVertex3d(-width, -height, -depth);
	glVertex3d(-width, height, -depth);
	glVertex3d(width, height, -depth);
	glVertex3d(width, -height, -depth);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glNormal3f(0, 0, 1);
	glVertex3d(-width, -height, depth);
	glVertex3d(-width, height, depth);
	glVertex3d(width, height, depth);
	glVertex3d(width, -height, depth);
	glEnd();
}
